class TruncationError(Exception):
    pass

class NotLoggedError(Exception):
    pass

class InvalidSenseValue(Exception):
    pass

class InvalidScaleValue(Exception):
    pass

class ModelError(Exception):
    pass
